# Andrey version 1.0.0
# AndreyParser.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of Andrey, a pedagogical implementation of Markov chains.
# Andrey is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Andrey is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Andrey.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Models
#   3.1) Models
#   3.2) Blocks & classes
#   3.3) States, events, parameters & rewards
#   3.4) Stochastic expressions
#   3.5) Transitions
#   3.6) Directives

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from S2MLParser import *
from AndreyModels import *
from AndreyLexer import *

# 2) Creator
# ----------

class AndreyParser(S2MLParser):
  def __init__(self, engine):
    S2MLParser.__init__(self, engine)

# 3) Models
# ---------

# 3.1) Models
# -----------

# 3.2) Blocks & classes
# ---------------------

  def _ParseBlockElement(self, container):
    element = None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None:
      return element
    elif token.GetType()==AndreyType.KEYWORD_STATE:
      element = self._ParseMarkovChainStateDeclaration(container)
    elif token.GetType()==AndreyType.KEYWORD_EVENT:
      element = self._ParseMarkovChainEventDeclaration(container)
    elif token.GetType()==AndreyType.KEYWORD_PARAMETER:
      element = self._ParseMarkovChainParameterDeclaration(container)
    elif token.GetType()==AndreyType.KEYWORD_REWARD:
      element = self._ParseMarkovChainRewardDeclaration(container)
    elif token.GetType()==AndreyType.KEYWORD_TRANSITION:
      element = self._ParseMarkovChainTransitionDeclaration(container)
    elif token.GetType()==AndreyType.KEYWORD_BLOCK:
      element = self._ParseBlock(container)
    elif token.GetType()==AndreyType.KEYWORD_EXTENDS:
      element = self._ParseExtendsDirective(container)
    elif token.GetType()==AndreyType.KEYWORD_EMBEDS:
      element = self._ParseExtendsDirective(container)
    elif token.GetType()==AndreyType.KEYWORD_CLONES:
      element = self._ParseClonesDirective(container)
    else:
      element = self._ParseClassInstance(container)
    return element

# 3.3) States, events, parameters & rewards
# -----------------------------------------

  def _ParseMarkovChainStateDeclaration(self, container):
    token = self._ParseGivenToken(AndreyType.KEYWORD_STATE)
    if not self._model.CanComposeState(container):
      self._RaiseParseError(token, "State declarations are not authorized here")
      return None
    container.SetSubType(AndreyType.MARKOV_CHAIN)
    states = []
    while True:
      path = self._ParsePath()
      if path==None:
        return None
      state = self._model.NewMarkovChainState(path, container)
      states.append(state)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()==AndreyType.TOKEN_COMMA:
        self._currentTokenIndex += 1
      else:
        break
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==AndreyType.TOKEN_LPAR:
      states = self._ParseAttributes(states)
      if states==None:
        return None
    token = self._ParseGivenToken(AndreyType.TOKEN_SEMICOLON)
    if token==None:
      return None
    return states

  def _ParseMarkovChainEventDeclaration(self, container):
    token = self._ParseGivenToken(AndreyType.KEYWORD_EVENT)
    if not self._model.CanComposeEvent(container):
      self._RaiseParseError(token, "Event declarations are not authorized here")
      return None
    container.SetSubType(AndreyType.MARKOV_CHAIN)
    events = []
    while True:
      path = self._ParsePath()
      if path==None:
        return None
      event = self._model.NewMarkovChainEvent(path, container)
      events.append(event)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()==AndreyType.TOKEN_COMMA:
        self._currentTokenIndex += 1
      else:
        break
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==AndreyType.TOKEN_LPAR:
      events = self._ParseAttributes(events)
      if events==None:
        return None
    token = self._ParseGivenToken(AndreyType.TOKEN_SEMICOLON)
    if token==None:
      return None
    return events

  def _ParseMarkovChainParameterDeclaration(self, container):
    token = self._ParseGivenToken(AndreyType.KEYWORD_PARAMETER)
    if not self._model.CanComposeParameter(container):
      self._RaiseParseError(token, "Parameter declarations are not authorized here")
      return None
    path = self._ParsePath()
    token = self._ParseGivenToken(AndreyType.TOKEN_EQUAL)
    if token==None:
      return None
    expression = self._ParseStochasticExpression()
    if expression==None:
      return None
    token = self._ParseGivenToken(AndreyType.TOKEN_SEMICOLON)
    if token==None:
      return None
    parameter = self._model.NewMarkovChainParameter(path, container)
    parameter.SetSubType(AndreyType.MARKOV_CHAIN_PARAMETER)
    parameter.SetDefinition(expression)
    return parameter

  def _ParseMarkovChainRewardDeclaration(self, container):
    token = self._ParseGivenToken(AndreyType.KEYWORD_REWARD)
    if not self._model.CanComposeReward(container):
      self._RaiseParseError(token, "Reward declarations are not authorized here")
      return None
    path = self._ParsePath()
    token = self._ParseGivenToken(AndreyType.TOKEN_EQUAL)
    if token==None:
      return None
    expression = self._ParseStochasticExpression()
    if expression==None:
      return None
    token = self._ParseGivenToken(AndreyType.TOKEN_SEMICOLON)
    if token==None:
      return None
    reward = self._model.NewMarkovChainParameter(path, container)
    reward.SetSubType(AndreyType.MARKOV_CHAIN_REWARD)
    reward.SetDefinition(expression)
    return reward

# 3.4) Stochastic expressions
# ---------------------------

# 3.5) Transitions
# ----------------

  def _ParseMarkovChainTransitionDeclaration(self, container):
    token = self._ParseGivenToken(AndreyType.KEYWORD_TRANSITION)
    if not self._model.CanComposeTransition(container):
      self._RaiseParseError(token, "Transitions declarations are not authorized here")
      return None
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or (token.GetType()!=AndreyType.TOKEN_IDENTIFIER and token.GetType()!=AndreyType.TOKEN_PATH):
        break
      self._currentTokenIndex += 1
      token = self._lexer.GetToken(self._currentTokenIndex)
      self._currentTokenIndex -= 1
      if token==None or token.GetType()!=AndreyType.TOKEN_COLON:
        break
      transition = self._ParseMarkovChainTransition(container)
      if transition==None:
        self._RecoverError(AndreyType.KEYWORD_END)
        return None
    return container

  def _ParseMarkovChainTransition(self, container):
    eventPath = self._ParsePath()
    if eventPath==None:
      return None
    token = self._ParseGivenToken(AndreyType.TOKEN_COLON)
    if token==None:
      return None
    sourceStatePath = self._ParsePath()
    if sourceStatePath==None:
      return None
    token = self._ParseGivenToken(AndreyType.TOKEN_RIGHT_ARROW)
    if token==None:
      return None
    targetStatePath = self._ParsePath()
    if targetStatePath==None:
      return None
    token = self._ParseGivenToken(AndreyType.TOKEN_SEMICOLON)
    if token==None:
      return None
    transition = self._model.NewMarkovChainTransition(eventPath, sourceStatePath, targetStatePath, container)
    return transition

# 3.6) Directives
# ---------------

  def _ParseInstance(self, container):
    directive = S2MLParser._ParseInstance(self, container)
    if directive==None:
      return None
    container.SetSubType(AndreyType.MARKOV_SYSTEM)
    return directive

  def _ParseClonesDirective(self, container):
    directive = S2MLParser._ParseClonesDirective(self, container)
    if directive==None:
      return None
    container.SetSubType(AndreyType.MARKOV_SYSTEM)
    return directive

  def _ParseEmbedsDirective(self, container):
    directive = S2MLParser._ParseEmbedsDirective(self, container)
    if directive==None:
      return None
    container.SetSubType(AndreyType.MARKOV_SYSTEM)
    return directive

