# S2MLCore version 1.0.0
# S2MLEngine.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Settings

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from S2MLStochasticExpressions import *
from S2MLLexer import *
from S2MLParser import *
from S2MLPrinter import *
from S2MLInterpreter import *

# 2) Creator
# ----------

class S2MLEngine:

  def __init__(self):
    self._lexer = None
    self._parser = None
    self._printer = None
    self._expressionManager = None
    self._flattener = None
    self._checker = None
    self._interpreter = None
    self._sourceModel = None
    self._targetModel = None
    self._targetBlock = None
    self._errors = 0
    self._errorFile = sys.stderr
    self._firstFreeMatrixIndex = 0
    self._matrix = None

# 3) Settings
# -----------

  def CreateTools(self):
    self._lexer = S2MLLexer(self)
    self._parser = S2MLParser(self)
    self._expressionManager = S2MLStochasticExpressionManager(self)
    self._printer = S2MLPrinter(self)
    self._flattener = S2MLFlattener(self)
    self._checker = S2MLChecker(self)
    self._interpreter = S2MLInterpreter(self)

  def GetLexer(self):
    return self._lexer
  def GetParser(self):
    return self._parser
  def GetExpressionManager(self):
    return self._expressionManager
  def GetPrinter(self):
    return self._printer
  def GetFlattener(self):
    return self._flattener
  def GetChecker(self):
    return self._checker
  def GetInterpreter(self):
    return self._interpreter
  def GetMatrix(self):
    return self._matrix

  def OpenSession(self):
    self.CreateTools()
    return 0

  def CloseSession(self):
    self._errorFile.flush()
    if self._errorFile!=sys.stderr:
      self._errorFile.close()
    return 0

  def SetErrorFile(self, fileName):
    try:
      file = open(fileName, "a")
    except:
      self.RaiseError('Unable to open file "%s"\n' % fileName)
      return 1
    self._errorFile.flush()
    if self._errorFile!=sys.stderr:
      self._errorFile.close()
    return 0

  def RaiseError(self, message):
    self._errors += 1
    self._errorFile.write("%s\n" % message)
    self._errorFile.flush()

# 4) Methods to import, instantiate and export models
# ---------------------------------------------------

  def ResetModel(self):
    self._sourceModel = S2MLModel()
    self._targetModel = None
    self._targetBlock = None
    return 0

  def ImportModel(self, fileName):
    self._targetModel = None
    self._targetBlock = None
    error = self._parser.ImportFile(self._sourceModel, fileName)
    if error!=0:
      self._sourceModel = None
    return error

  def LoadScript(self, fileName):
    self._interpreter.LoadScript(fileName)

  def ExportModel(self, fileName, mode="w"):
    if self._sourceModel==None:
      self.RaiseError("No model to export")
      return 1
    error = self._printer.ExportModel(self._sourceModel, fileName, mode)
    return error

  def InstantiateModel(self):
    if self._sourceModel==None:
      self.RaiseError("No model to check")
      return 1
    self._targetModel = self._flattener.InstantiateAndFlattenModel(self._sourceModel)
    self._targetBlock = None
    if self._targetModel==None:
      return 1
    error = self._checker.CheckModel(self._targetModel)
    if error!=0:
      self._targetModel = None
      return error
    self._targetBlock = self._targetModel.GetBlock(0)
    return 0

  def ExportInstantiatedModel(self, fileName, mode="w"):
    if self._targetModel==None:
      self.RaiseError("No model to export")
      return 1
    self._printer.ExportModel(self._targetModel, fileName, mode)
    return 0

  def SetTargetBlock(self, blockName):
    if self._targetModel==None:
      self.RaiseError("No instantiated model")
      return 1
    self._targetBlock = self._targetModel.LookForBlock(blockName)
    if self._targetBlock==None:
      self.RaiseError("No block %s in instantiated model" % blockName)
      return 1
    return 0

