# S2MLCore version 1.0.0
# S2MLInterpreter.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Scripts
# 4) Methods to parse the command line

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from S2MLMatrices import *

# 2) Creator
# ----------

class S2MLInterpreter:
  def __init__(self, engine):
    from S2MLEngine import S2MLEngine
    self._engine = engine
    self._printer = engine.GetPrinter()
    self._inputFileName = None
    self._inputFile = sys.stdin
    self._errors = 0
    self._line = None
    self._lineNumber = 0
    self._arguments = []
    self._argumentIndex = 0
    self._listLevel = 0

  def _RaiseError(self, message):
    self._errors += 1
    self._engine.RaiseError(message)

# 3) Scripts
# ----------

  def LoadScript(self, fileName):
    try:
      file = open(fileName, "r")
    except:
      self._RaiseError('Unable to open file "%s"\n' % fileName)
      return 1
    self._errors = self.ParseFile(fileName, file)
    file.close()
    return self._errors

  def ParseFile(self, fileName, file):
    previousInputFileName = self._inputFileName
    previousFile = self._inputFile
    previousLine = self._line
    previousLineNumber = self._lineNumber
    self._inputFileName = fileName
    self._inputFile = file
    self._line = None
    self._lineNumber = 0
    self._ParseCommands()
    self._inputFileName = previousInputFileName
    self._inputFile = previousFile
    self._line = previousLine
    self._lineNumber = previousLineNumber

  def _ParseCommands(self):
    for self._line in self._inputFile:
      self._ParseCommandLine()
      if self._arguments!=[]:
        self._ParseCommand()

  def _ParseCommand(self):
    commandName = self.GetArgument()
    if commandName=="":
      print("uh?")
    elif commandName=="instantiate":
      self._errors += self._ParseCommandInstantiate()
    elif commandName=="load":
      self._errors += self._ParseCommandLoad()
    elif commandName=="print":
      self._errors += self._ParseCommandPrint()
    elif commandName=="set":
      self._errors += self._ParseCommandSet()
    else:
      self._RaiseError("line %d\nUnknow command %s\n" % (self._lineNumber, commandName))

  # load model fileName
  # load script fileName
  def _ParseCommandLoad(self):
    what = self.GetArgument()
    if what!="model" and what!="script":
      self._RaiseError("line %d\nIncorrect argument %s\n" % (self._lineNumber, what))
      return 1
    fileName = self.GetArgument()
    if fileName==None:
      self._RaiseError("line %d\nA file name was expected\n" % self._lineNumber)
      return 1
    if what=="model":
      return self._engine.ImportModel(fileName)
    if what=="script":
      return self._engine.LoadScript(fileName)
    return 1 # impossible

  # instantiate model
  def _ParseCommandInstantiate(self):
    what = self.GetArgument()
    if what!="model":
      self._RaiseError("line %d\nIncorrect argument %s\n" % (self._lineNumber, what))
      return 1
    if what=="model":
      return self._engine.InstantiateModel()
    return 1 # impossible

  # print model fileName [mode=(append|write)]
  # print instantiated-model fileName [mode=(append|write)]
  def _ParseCommandPrint(self):
    what = self.GetArgument()
    if what!="model" and what!="instantiated-model":
      self._RaiseError("line %d\nIncorrect argument %s\n" % (self._lineNumber, what))
      return 1
    fileName = self.GetArgument()
    if fileName==None:
      self._RaiseError("line %d\nA file name was expected\n" % self._lineNumber)
      return 1
    mode = self.GetModeOption("w")
    if mode==None:
      return 1
    if what=="model":
      return self._engine.ExportModel(fileName, mode=mode)
    if what=="instantiated-model":
      return self._engine.ExportInstantiatedModel(fileName, mode=mode)
    return 1 # impossible

  def _ParseCommandCompute(self):
    what = self.GetArgument()
    self._RaiseError("line %d\nSyntax error\n" % self._lineNumber)
    return 1

  # set block blockName
  def _ParseCommandSet(self):
    what = self.GetArgument()
    if what!="block":
      self._RaiseError("line %d\nIncorrect argument %s\n" % (self._lineNumber, what))
      return 1
    blockName = self.GetArgument()
    if blockName==None:
      self._RaiseError("line %d\nA block name was expected\n" % self._lineNumber)
      return 1
    if what=="block":
      return self._engine.SetTargetBlock(blockName)
    return 1 # impossible

# 4) Methods to parse the command line
# ------------------------------------

  def _ParseCommandLine(self):
    self._lineNumber += 1
    self._line.rstrip("\r\n")
    self._line.rstrip("\n")
    self._arguments = []
    state = 1
    column = 0
    while True:
      if state==1:
        if column>=len(self._line):
          if self._listLevel==0:
            break
          else:
            self._line = self._inputFile.readline()
            self._lineNumber += 1
            self._line.rstrip("\r\n")
            self._line.rstrip("\n")
            column = 0
        else:
          c = self._line[column]
          if c.isspace():
            pass
          elif c=="#":
            break
          elif c=="=":
            self._arguments.append("=")
          elif c==",":
            self._arguments.append(",")
          elif c=="[":
            self._arguments.append("[")
            self._listLevel += 1
          elif c=="]":
            self._arguments.append("]")
            self._listLevel -= 1
          elif c=="'":
            start = column+1
            state = 3
          elif c=='"':
            start = column+1
            state = 4
          elif c=="\\":
            self._line = self._inputFile.readline()
            self._lineNumber += 1
            self._line.rstrip("\r\n")
            self._line.rstrip("\n")
            column = 0
          else:
            start = column
            state = 2
          column += 1
      elif state==2:
        if column>=len(self._line):
          argument = self._line[start:column]
          self._arguments.append(argument)
          break
        else:
          c = self._line[column]
          if c.isspace() or c=="#" or c=="=" or c=="'" or c=='"' or c=="," or c=="[" or c=="]":
            argument = self._line[start:column]
            self._arguments.append(argument)
            state = 1
          else:
            column += 1
      elif state==3:
        if column>=len(self._line):
          self._RaiseError("Line %d, unterminated string" % self._lineNumber)
          self._arguments = []
          break
        else:
          c = self._line[column]
          if c=="'":
            argument = self._line[start:column]
            self._arguments.append(argument)
            state = 1
          column += 1
      elif state==4:
        if column>=len(self._line):
          self._RaiseError("Line %d, unterminated string" % self._lineNumber)
          self._arguments = []
          break
        else:
          c = self._line[column]
          if c=='"':
            argument = self._line[start:column]
            self._arguments.append(argument)
            state = 1
          column += 1
      self._argumentIndex = 0
    return

  def LookUpArgument(self):
    if self._argumentIndex>=len(self._arguments):
      return None
    return self._arguments[self._argumentIndex]

  def GetArgument(self):
    if self._argumentIndex>=len(self._arguments):
      return None
    argument = self._arguments[self._argumentIndex]
    self._argumentIndex += 1
    return argument

  def GetStringOption(self, name, defaultValue):
    argument = self.LookUpArgument()
    if argument!=name:
      return defaultValue
    if self._argumentIndex+2>=len(self._arguments):
      self._RaiseError("Line %d. Wrong number of arguments" % self._lineNumber)
      return None
    if self._arguments[self._argumentIndex+1]!="=":
      self._RaiseError("Line %d. A '=' was expected" % self._lineNumber)
      return None
    option = self._arguments[self._argumentIndex+2]
    self._argumentIndex += 3
    return option

  def GetBoolOption(self, name, defaultValue):
    argument = self.LookUpArgument()
    if argument!=name:
      return defaultValue
    if self._argumentIndex+2>=len(self._arguments):
      self._RaiseError("Line %d. Wrong number of arguments" % self._lineNumber)
      return None
    if self._arguments[self._argumentIndex+1]!="=":
      self._RaiseError("Line %d. A '=' was expected" % self._lineNumber)
      return None
    option = self._arguments[self._argumentIndex+2].lower()
    if option!="true" and option!="false":
      self._RaiseError("Line %d. A Boolean (true or false) was expected" % self._lineNumber)
      return None
    self._argumentIndex += 3
    return option=="true"

  def GetIntOption(self, name, defaultValue):
    argument = self.LookUpArgument()
    if argument!=name:
      return defaultValue
    if self._argumentIndex+2>=len(self._arguments):
      self._RaiseError("Line %d. Wrong number of arguments" % self._lineNumber)
      return None
    if self._arguments[self._argumentIndex+1]!="=":
      self._RaiseError("Line %d. A '=' was expected" % self._lineNumber)
      return None
    string = self._arguments[self._argumentIndex+2]
    if not self.IsInt(string):
      self._RaiseError("Line %d. An integer was expected" % self._lineNumber)
      return None
    option = int(string)
    self._argumentIndex += 3
    return option

  def GetFloatOption(self, name, defaultValue):
    argument = self.LookUpArgument()
    if argument!=name:
      return defaultValue
    if self._argumentIndex+2>=len(self._arguments):
      self._RaiseError("Line %d. Wrong number of arguments" % self._lineNumber)
      return None
    if self._arguments[self._argumentIndex+1]!="=":
      self._RaiseError("Line %d. A '=' was expected" % self._lineNumber)
      return None
    string = self._arguments[self._argumentIndex+2]
    if not self.IsFloat(string):
      self._RaiseError("Line %d. A float was expected" % self._lineNumber)
      return None
    option = float(string)
    self._argumentIndex += 3
    return option

  def GetModeOption(self, defaultValue):
    argument = self.LookUpArgument()
    if argument!="mode":
      return defaultValue
    if self._argumentIndex+2>=len(self._arguments):
      self._RaiseError("Line %d. Wrong number of arguments" % self._lineNumber)
      return None
    if self._arguments[self._argumentIndex+1]!="=":
      self._RaiseError("Line %d. A '=' was expected" % self._lineNumber)
      return None
    mode = self._arguments[self._argumentIndex+2]
    if mode=="write":
      mode = "w"
    if mode=="append":
      mode = "a"
    if mode!="w" and mode!="a":
      return None
    self._argumentIndex += 3
    return mode

  def GetIntList(self):
    items = []
    argument = self.LookUpArgument()
    if argument=="[":
      argument = self.GetArgument()
      items = []
      while True:
        item = self.GetInt()
        if item==None:
          return None
        items.append(item)
        argument = self.GetArgument()
        if argument=="]":
          break
        elif argument==",":
          pass
        else:
          self._RaiseError("Line %d. Syntax error" % self._lineNumber)
          return None
    else:
      item = self.GetInt()
      if item==None:
        return None
      items = [item]
    return items

  def GetFloatList(self):
    items = []
    argument = self.LookUpArgument()
    if argument=="[":
      argument = self.GetArgument()
      items = []
      while True:
        item = self.GetFloat()
        if item==None:
          return None
        items.append(item)
        argument = self.GetArgument()
        if argument=="]":
          break
        elif argument==",":
          pass
        else:
          self._RaiseError("Line %d. Syntax error" % self._lineNumber)
          return None
    else:
      item = self.GetFloat()
      if item==None:
        return None
      items = [item]
    return items

  def GetInt(self):
    argument = self.GetArgument()
    if argument==None:
      return None
    if not self.IsInt(argument):
      self._RaiseError("Line %d. An int was expected" % self._lineNumber)
      return None
    return int(argument)

  def GetFloat(self):
    argument = self.GetArgument()
    if argument==None:
      return None
    if not self.IsFloat(argument):
      self._RaiseError("Line %d. A float was expected" % self._lineNumber)
      return None
    return float(argument)

  def IsInt(self, item):
    try:
      value = int(item)
      return True
    except ValueError:
      return False

  def IsFloat(self, item):
    try:
      value = float(item)
      return True
    except ValueError:
      return False
