# S2MLCore version 1.0.0
# S2MLModels.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) S2ML constructs
#  2.1) Domains
#  2.2) Ports
#  2.3) Base blocks
#  2.4) Blocks
#  2.5) Classes
#  2.6) Instances
#  2.7) Extends directive
#  2.8) Clones directive
#  2.9) Embeds directive
#  2.10) Models

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from S2MLType import *
from S2MLMetaModel import *
from S2MLStochasticExpressions import *

# 2) S2ML constructs
# ==================

# 2.1) Domains
# ------------

class S2MLDomain(S2MLElement):
  def __init__(self, identifier, container):
    S2MLElement.__init__(self, S2MLType.DOMAIN, container)
    path = S2MLPath.NewEmptyPath()
    path.AppendIdentifier(identifier)
    self.SetPath(path)
    self.Rename(identifier)
    self._constants = []

  def GetSize(self):
    return len(self._constants)
  def GetConstants(self):
    return self._constants
  def GetConstant(self, i):
    if i<0 or i>=self.GetSize():
      return None
    return self._constants[i]
  def AppendConstant(self, constant):
    self._constants.append(constant)
  def AppendConstants(self, constants):
    self._constants.extend(constants)
  def Contains(self, constant):
    return constant in self._constants

# 2.2) Ports
# ----------

class S2MLPort(S2MLElement):
  def __init__(self, type, domainIdentifier, path, container):
    S2MLElement.__init__(self, type, container)
    self._domainIdentifier = domainIdentifier
    self._domain = None
    self.SetPath(path)

  def GetDomainIdentifier(self):
    return self._domainIdentifier
  def SetDomainIndentifier(self, identifier):
    self._domainIdentifier = identifier

  def GetDomain(self):
    return self._domain
  def SetDomain(self, domain):
    self._domain = domain

# 2.3) Base blocks
# ----------------

class S2MLBaseBlock(S2MLContainer):
  def __init__(self, type, identifier, container):
    S2MLContainer.__init__(self, type, container)
    path = S2MLPath.NewEmptyPath()
    path.AppendIdentifier(identifier)
    self.SetPath(path)
    self.Rename(identifier)

  def LookForBlock(self, identifier):
    element = self.LookForIdentifiedElement(identifier)
    if element==None or not S2MLModel.IsBlock(element):
      return None
    return element
  def CollectBlocks(self, recursive = False):
    blocks = []
    self.CollectElements(S2MLModel.IsBlock, recursive, blocks)
    return blocks
  def GetNumberOfBlocks(self):
    return len(self.CollectBlocks())
  def GetBlock(self, index):
    blocks = self.CollectBlocks()
    if index<0 or index>=len(blocks):
      return None
    return blocks[index]

# 2.4) Blocks
# -----------

class S2MLBlock(S2MLBaseBlock):
  def __init__(self, identifier, container):
    S2MLBaseBlock.__init__(self, S2MLType.BLOCK, identifier, container)

# 2.5) Classes
# -----------

class S2MLClass(S2MLBaseBlock):
  def __init__(self, identifier, container):
    S2MLBaseBlock.__init__(self, S2MLType.CLASS, identifier, container)

# 2.6) Instances
# --------------

class S2MLInstance(S2MLBaseBlock):
  def __init__(self, identifier, classIdentifier, container):
    S2MLBaseBlock.__init__(self, S2MLType.INSTANCE, identifier, container)
    self._classIdentifier = classIdentifier

  def GetClassIdentifier(self):
    return self._classIdentifier
  def SetClassIdentifier(self, classIdentifier):
    self._classIdentifier = classIdentifier

# 2.7) Extends directive
# ----------------------

class S2MLExtendsDirective(S2MLElement):
  def __init__(self, classIdentifier, container):
    S2MLElement.__init__(self, S2MLType.EXTENDS, container)
    self._classIdentifier = classIdentifier

  def GetClassIdentifier(self):
    return self._classIdentifier
  def SetClassIdentifier(self, classIdentifier):
    self._classIdentifier = classIdentifier

# 2.8) ClonesDirective
# --------------------

class S2MLClonesDirective(S2MLBaseBlock):
  def __init__(self, identifier, pathToClonedElement, container):
    S2MLBaseBlock.__init__(self, S2MLType.CLONES, identifier, container)
    self._pathToClonedElement = pathToClonedElement

  def GetPathToClonedElement(self):
    return self._pathToClonedElement
  def SetPathToClonedElement(self, pathToClonedElement):
    self._pathToClonedElement = pathToClonedElement 

# 2.9) EmbedsDirective
# --------------------

class S2MLEmbedsDirective(S2MLBaseBlock):
  def __init__(self, identifier, pathToEmbeddedElement, container):
    S2MLBaseBlock.__init__(self, S2MLType.EMBEDS, identifier, container)
    self._pathToEmbeddedElement = pathToEmbeddedElement

# 2.10) Models
# ------------

class S2MLModel(S2MLBaseBlock):

  # Creator
  # -------

  def __init__(self):
    S2MLBaseBlock.__init__(self, S2MLType.MODEL, "___MODEL__", None)
    self._universe = S2MLUniverse()

  # Methods to manage the universe
  # ------------------------------
  def GetNone(self):
    return self._universe.GetNone()
  def NewNone(self):
    return self._universe.NewNone()

  def GetBoolean(self, value):
    return self._universe.GetBoolean(value)
  def NewBoolean(self):
    return self._universe.NewBoolean(value)

  def GetInteger(self, value):
    return self._universe.GetInteger(value)
  def NewInteger(self):
    return self._universe.NewInteger(value)

  def GetReal(self, value):
    return self._universe.GetReal(value)
  def NewReal(self):
    return self._universe.NewReal(value)

  def GetSymbol(self, value):
    return self._universe.GetSymbol(value)
  def NewSymbol(self, value):
    return self._universe.NewSymbol(value)

  def GetString(self, value):
    return self._universe.GetString(value)
  def NewString(self, value):
    return self._universe.NewString(value)

  # Methods to test types
  # ---------------------

  @staticmethod
  def IsDomain(element):
    return element.GetType()==S2MLType.DOMAIN
  @staticmethod
  def IsBlock(element):
    return element.GetType()==S2MLType.BLOCK
  @staticmethod
  def IsClass(element):
    return element.GetType()==S2MLType.CLASS

  @staticmethod
  def IsParameter(element):
    return element.GetType()==S2MLType.PARAMETER

  @staticmethod
  def IsModel(element):
    return element.GetType()==S2MLType.MODEL

  def IsValidReferredElementInStochasticExpression(self, element):
    return element.GetType()==S2MLType.PARAMETER

  # Factory of elements (to be derived)
  # ----------------------------------

  def NewDomain(self, identifier, container):
    return S2MLDomain(identifier, container)

  def NewPort(self, type, domainIdentifier, path, container):
    return S2MLPort(type, domainIdentifier, path, container)

  def NewBlock(self, identifier, container):
    return S2MLBlock(identifier, container)

  def NewClass(self, identifier, container):
    return S2MLClass(identifier, container)

  def NewInstance(self, identifier, classIdentifier, container):
    return S2MLInstance(identifier, classIdentifier, container)

  def NewExtendsDirective(self, classIdentifier, container):
    return S2MLExtendsDirective(classIdentifier, container)

  def NewClonesDirective(self, identifier, pathToClonedElement, container):
    return S2MLClonesDirective(identifier, pathToClonedElement, container)

  def NewEmbedsDirective(self, identifier, pathToEmbeddedElement, container):
    return S2MLEmbedsDirective(identifier, pathToEmbeddedElement, container)

  def NewParameter(self, path, container):
    return S2MLParameter(path, container)

  # Management of classes
  # ---------------------

  def LookForClass(self, name):
    element = self.LookForIdentifiedElement(name)
    if element==None or not S2MLModel.IsClass(element):
      return None
    return element

  def CollectClasses(self):
    classes = []
    self.CollectElements(S2MLModel.IsClass, False, classes)
    return classes

  def GetNumberOfClasses(self):
    return len(self.CollectClasses())

  # Methods for the parser
  # ----------------------

  def CanComposeDomain(self, container):
    return S2MLModel.IsModel(container)

  def CanComposePort(self, container):
    return True

  def CanComposeBlock(self, container):
    return True

  def CanComposeClass(self, container):
    return S2MLModel.IsModel(container)

  def CanComposeInstance(self, container):
    return True

  def CanComposeClonesDirective(self, container):
    return S2MLModel.IsBlock(container) or S2MLModel.IsClass(container)

  def CanComposeExtendsDirective(self, container):
    return S2MLModel.IsBlock(container) or S2MLModel.IsClass(container)

  def CanComposeEmbedsDirective(self, container):
    return S2MLModel.IsBlock(container) or S2MLModel.IsClass(container)

  def CanComposeParameter(self, container):
    return True




