# S2MLCore version 1.0.0
# S2MLParser.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Models
#   3.1) Models
#   3.2) Blocks and classes
#   3.3) Directives
#   3.4) Attributes
#   3.5) Parameters and stochastic expressions
# 4) Auxiliary functions

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from S2MLModels import *
from S2MLLexer import *

# 2) Creator
# ----------

class S2MLParser:
  def __init__(self, engine):
    from S2MLEngine import S2MLEngine
    self._engine = engine
    self._lexer = self._engine.GetLexer()
    self._currentTokenIndex = 0
    self._model = None

# 3) Models
# ---------

# 3.1) Models
# -----------

  def ImportFile(self, model, fileName):
    try:
      file = open(fileName, "r")
    except:
      self._engine.RaiseError('Unable to open file "%s"' % fileName)
      return 1
    self.ParseFile(model, fileName, file)
    file.close()
    return self._errors

  def ParseFile(self, model, fileName, file):
    self._errors = self._lexer.Initialize(fileName, file)
    if self._errors!=0:
      return
    self._ParseModel(model)
    self._lexer.Finalize()
    return self._errors

  def _ParseModel(self, model):
    self._model = model
    self._currentTokenIndex = 0
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None:
        break
      elif token.GetType()==S2MLType.KEYWORD_DOMAIN:
        self._ParseDomain(self._model)
      elif token.GetType()==S2MLType.KEYWORD_BLOCK:
        self._ParseBlock(self._model)
      elif token.GetType()==S2MLType.KEYWORD_CLASS:
        self._ParseClass(self._model)
      else:
        self._RaiseParseError(token, "A declaration of class or block was expected ")
        self._currentTokenIndex += 1

# 3.2) Blocks & classes
# ---------------------

  # This method can possibly be derived
  def _ParseDomain(self, container):
    token = self._ParseGivenToken(S2MLType.KEYWORD_DOMAIN)
    if not self._model.CanComposeDomain(container):
      self._RaiseParseError(token, "Domains are not authorized here")
      self._RecoverError(S2MLType.TOKEN_RBRACE)
      return None
    identifier = self._ParseIdentifier()
    if identifier==None:
      self._RecoverError(S2MLType.TOKEN_RBRACE)
      return None
    domain = self._model.NewDomain(identifier, container)
    token = self._ParseGivenToken(S2MLType.TOKEN_LBRACE)
    if token==None:
      return None
    while True:
      symbol = self._ParseIdentifier()
      if symbol==None:
        return None
      constant = self._model.NewSymbol(symbol)
      domain.AppendConstant(constant)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or token.GetType()!=S2MLType.TOKEN_COMMA:
        break
      self._currentTokenIndex += 1
    token = self._ParseGivenToken(S2MLType.TOKEN_RBRACE)
    if token==None:
      return None
    return domain

  # This method can possibly be derived
  def _ParseBlock(self, container):
    token = self._ParseGivenToken(S2MLType.KEYWORD_BLOCK)
    if not self._model.CanComposeBlock(container):
      self._RaiseParseError(token, "Blocks are not authorized here")
      self._RecoverError(S2MLType.KEYWORD_END)
      return None
    identifier = self._ParseIdentifier()
    if identifier==None:
      self._RecoverError(S2MLType.KEYWORD_END)
      return None
    blockDeclaration = self._model.NewBlock(identifier, container)
    return self._ParseBlockBody(blockDeclaration)

  # This method can possibly be derived
  def _ParseClass(self, container):
    token = self._ParseGivenToken(S2MLType.KEYWORD_CLASS)
    if not self._model.CanComposeClass(container):
      self._RaiseParseError(token, "Classes are not authorized here")
      self._RecoverError(S2MLType.KEYWORD_END)
      return None
    identifier = self._ParseIdentifier()
    if identifier==None:
      self._RecoverError(S2MLType.KEYWORD_END)
      return None
    classDeclaration = self._model.NewClass(identifier, container)
    return self._ParseBlockBody(classDeclaration)

  def _ParseBlockBody(self, container):
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or token.GetType()==S2MLType.KEYWORD_END:
        break
      blockElement = self._ParseBlockElement(container)
      if blockElement==None:
        self._RecoverError(S2MLType.KEYWORD_END)
        return None
    token = self._ParseGivenToken(S2MLType.KEYWORD_END)
    if token==None:
      return None
    return container

  # This method must be derived
  def _ParseBlockElement(self, container):
    element = None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None:
      return None
    elif token.GetType()==S2MLType.KEYWORD_PARAMETER:
      element = self._ParseParameter(container)
    elif token.GetType()==S2MLType.KEYWORD_BLOCK:
      element = self._ParseBlock(container)
    elif token.GetType()==S2MLType.KEYWORD_EXTENDS:
      element = self._ParseExtendsDirective(container)
    elif token.GetType()==S2MLType.KEYWORD_EMBEDS:
      element = self._ParseExtendsDirective(container)
    elif token.GetType()==S2MLType.KEYWORD_CLONES:
      element = self._ParseClonesDirective(container)
    else:
      instantiatedElementIdentifier = self._ParseIdentifier()
      if instantiatedElementIdentifier==None:
        return None
      instantiatedElement = self._model.LookForIdentifiedElement(instantiatedElementIdentifier)
      if instantiatedElement==None:
        self._RaiseParseError(token, "Undeclared domain or class %s" % instantiatedElementIdentifier)
        return None
      self._currentTokenIndex -= 1
      if instantiatedElement.GetType()==S2MLType.DOMAIN:
        element = self._ParsePortDeclaration(container)
      else:
        element = self._ParseClassInstance(container)
    return element

# 3.3) Directives
# ---------------

  def _ParsePortDeclaration(self, container):
    if not self._model.CanComposePort(container):
      token = self._lexer.GetToken(self._currentTokenIndex)
      self._RaiseParseError(token, "Port declarations are not authorized here")
      return None
    domainIdentifier = self._ParseIdentifier()
    if domainIdentifier==None:
      return None
    domain = self._model.LookForIdentifiedElement(domainIdentifier)
    if domain==None:
      self._RaiseParseError(token, "Undeclared domain %s" % domainIdentifier)
      return None
    ports = []
    while True:
      path = self._ParsePath()
      if path==None:
        return None
      port = self._model.NewPort(S2MLType.PORT, domainIdentifier, path, container)
      ports.append(port)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()==CheckyType.TOKEN_COMMA:
        self._currentTokenIndex += 1
      else:
        break
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==CheckyType.TOKEN_LPAR:
      ports = self._ParseAttributes(ports)
      if ports==None:
        return None
    token = self._ParseGivenToken(CheckyType.TOKEN_SEMICOLON)
    if token==None:
      return None
    return ports

  def _ParseClassInstance(self, container):
    if not self._model.CanComposeInstance(container):
      token = self._lexer.GetToken(self._currentTokenIndex)
      self._RaiseParseError(token, "Instances are not authorized here")
      return None
    classIdentifier = self._ParseIdentifier()
    if classIdentifier==None:
      return None
    instantiatedClass = self._model.LookForIdentifiedElement(classIdentifier)
    if instantiatedClass==None:
      self._RaiseParseError(token, "Undeclared class %s" % classIdentifier)
      return None
    identifier = self._ParseIdentifier()
    if identifier==None:
      return None
    directive = self._model.NewInstance(identifier, classIdentifier, container)
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==S2MLType.TOKEN_SEMICOLON:
      token = self._ParseGivenToken(S2MLType.TOKEN_SEMICOLON)
    else:
      directive = self._ParseBlockBody(directive)
      if directive==None:
        return None
    return directive

  def _ParseClonesDirective(self, container):
    if not self._model.CanComposeClonesDirective(container):
      self._RaiseParseError(token, "Clones directives are not authorized here")
      return None
    token = self._ParseGivenToken(S2MLType.KEYWORD_CLONES)
    pathToClonedElement = self._ParsePath()
    if pathToClonedElement==None:
      return None
    token = self._ParseGivenToken(S2MLType.KEYWORD_AS)
    if token==None:
      return None
    identifier = self._ParseIdentifier()
    if identifier==None:
      return None
    directive = self._model.NewClonesDirective(identifier, pathToClonedElement, container)
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==S2MLType.TOKEN_SEMICOLON:
      token = self._ParseGivenToken(S2MLType.TOKEN_SEMICOLON)
    else:
      directive = self._ParseBlockBody(directive)
      if directive==None:
        return None
    return directive

  def _ParseExtendsDirective(self, container):
    if not self._model.CanComposeExtendsDirective(container):
      self._RaiseParseError(token, "Extends directives are not authorized here")
      return None
    token = self._ParseGivenToken(S2MLType.KEYWORD_EXTENDS)
    classIdentifier = self._ParseIdentifier()
    if classIdentifier==None:
      return None
    token = self._ParseGivenToken(S2MLType.TOKEN_SEMICOLON)
    if token==None:
      return None
    directive = self._model.NewExtendsDirective(classIdentifier, container)
    return directive

  def _ParseEmbedsDirective(self, container):
    if not self._model.CanComposeEmbedsDirective(container):
      self._RaiseParseError(token, "Embeds directives are not authorized here")
      return None
    token = self._ParseGivenToken(S2MLType.KEYWORD_EMBEDS)
    pathToEmbeddedElement = self._ParsePath()
    if pathToEmbeddedElement==None:
      return None
    token = self._ParseGivenToken(S2MLType.KEYWORD_AS)
    if token==None:
      return None
    identifier = self._ParseIdentifier()
    if identifier==None:
      return None
    directive = self._model.NewEmbedsDirective(identifier, pathToEmbeddedElement, container)
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==S2MLLexer.SEMICOLON:
      token = self._ParseGivenToken(S2MLType.TOKEN_SEMICOLON)
    else:
      directive = self._ParseBlockBody(directive)
      if directive==None:
        return None
    return directive

# 3.4) Attributes
# ---------------

  def _ParseAttributes(self, elements):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()!=S2MLType.TOKEN_LPAR:
      return elements
    self._currentTokenIndex += 1
    while True:
      identifier = self._ParseIdentifier()
      if identifier==None:
        return None
      token = self._ParseGivenToken(S2MLType.TOKEN_EQUAL)
      if token==None:
        return None
      expression = self._ParseStochasticExpression()
      if expression==None:
        return None
      for element in elements:
        element.SetAttribute(identifier, expression)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()==S2MLType.TOKEN_RPAR:
        break
      token = self._ParseGivenToken(S2MLType.TOKEN_COMMA)
      if token==None:
        return None
    token = self._ParseGivenToken(S2MLType.TOKEN_RPAR)
    if token==None:
      return None
    return elements

# 3.5) Parameters and stochastic expressions
# ------------------------------------------

  def _ParseParameter(self, container):
    token = self._ParseGivenToken(S2MLType.KEYWORD_PARAMETER)
    if not self._model.CanComposeParameter(container):
      self._RaiseParseError(token, "Parameter declarations are not authorized here")
      return None
    path = self._ParsePath()
    if path==None:
      return None
    token = self._ParseGivenToken(S2MLType.TOKEN_EQUAL)
    if token==None:
      return None
    expression = self._ParseStochasticExpression()
    if expression==None:
      return None
    token = self._ParseGivenToken(S2MLType.TOKEN_SEMICOLON)
    if token==None:
      return None
    parameter = self._model.NewParameter(path, container)
    parameter.SetDefinition(expression)
    return parameter

  def _ParseStochasticExpression(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==S2MLType.TOKEN_IDENTIFIER:
      identifier = self._ParseIdentifier()
      symbol = self._model.GetSymbol(identifier)
      if symbol==None:
        path = S2MLPath.NewEmptyPath()
        path.AppendIdentifier(identifier)
        expression = S2MLStochasticExpression.NewReference(path)
      else:
        expression = S2MLStochasticExpression.NewConstant(S2MLConstant.Copy(symbol))
      return expression
    if token.GetType()==S2MLType.TOKEN_PATH:
      path = self._ParsePath()
      expression = S2MLStochasticExpression.NewReference(path)
      return expression
    if token.GetType()==S2MLType.TOKEN_NUMBER:
      token = self._ParseGivenToken(S2MLType.TOKEN_NUMBER)
      expression = S2MLStochasticExpression.NewConstant(S2MLConstant.NewReal(S2MLReal.New(token.GetString())))
      return expression
    if token.GetType()==S2MLType.KEYWORD_FALSE:
      token = self._ParseGivenToken(S2MLType.KEYWORD_FALSE)
      expression = S2MLStochasticExpression.NewConstant(S2MLConstant.NewBoolean(False))
      return expression
    if token.GetType()==S2MLType.KEYWORD_TRUE:
      token = self._ParseGivenToken(S2MLType.KEYWORD_TRUE)
      expression = S2MLStochasticExpression.NewConstant(S2MLConstant.NewBoolean(True))
      return expression
    token = self._ParseGivenToken(S2MLType.TOKEN_LPAR)
    if token==None:
      return None
    token = self._lexer.GetToken(self._currentTokenIndex)
    operator = self._engine.GetExpressionManager().GetOperatorOfKeywordCode(token.GetType())
    if operator==None:
      self._RaiseParseError(token, "Unknown operator")
      return None
    self._currentTokenIndex += 1
    children = self._ParseStochasticExpressionList()
    if children==None:
      return None
    if len(children)<operator.GetMinimumNumberOfArguments():
      self._RaiseParseError(token, "Too few arguments")
      return None
    if operator.GetMaximumNumberOfArguments()!=S2MLStochasticOperator.MANY and len(children)>operator.GetMaximumNumberOfArguments():
      self._RaiseParseError(token, "Too many arguments")
      return None
    token = self._ParseGivenToken(S2MLType.TOKEN_RPAR)
    if token==None:
      return None
    return S2MLStochasticExpression.NewOperation(operator.GetType(), children)

  def _ParseStochasticExpressionList(self):
    expressions = []
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()==S2MLType.TOKEN_RPAR:
        break
      expression = self._ParseStochasticExpression()
      if expression==None:
        return None
      expressions.append(expression)
    return expressions

# 4) Auxiliary functions
# ----------------------

  def _ParseIdentifier(self):
    token = self._ParseGivenToken(S2MLType.TOKEN_IDENTIFIER)
    if token==None:
      return None
    return token.GetString()

  def _ParsePath(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None or (token.GetType()!=S2MLType.TOKEN_IDENTIFIER and token.GetType()!=S2MLType.TOKEN_PATH):
      self._RaiseParseError(token, "A path was expected")
      return None
    self._currentTokenIndex += 1
    return S2MLPath.FromString(token.GetString())

  def _ParseToken(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None:
      self._RaiseParseError(token, "Unexpected end of file")
      return None
    self._currentTokenIndex += 1
    return token

  def _ParseGivenToken(self, tokenType):
    token = self._ParseToken()
    if token==None:
      return None
    if token.GetType()!=tokenType:
      self._RaiseParseError(token, "A %s was expected" % self._lexer.GetStringOfCode(tokenType))
      return None
    return token

  def _RecoverError(self, tokenType):
    currentTokenIndex = self._currentTokenIndex
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None:
        break
      self._currentTokenIndex += 1
      if token.GetType()==tokenType:
        break
    if currentTokenIndex==self._currentTokenIndex:
      sys.exit()

  def _RaiseParseError(self, token, message):
    self._errors += 1
    fullMessage = ""
    if token!=None:
      fullMessage += "Parse error line %d (%d, %s):\n" % (token.GetLineNumber(), token.GetColumnNumber(), token.GetString())
    fullMessage += message
    self._engine.RaiseError(fullMessage)

